#include "fpsplayer.h"

FPSPlayer* FPSPlayer::thePlayer = NULL;
unsigned int PlayerCharacter::frameListenerPriority = 1;

class MyTrigger : public Trigger
{
public:
	MyTrigger(const NxString& meshIdentifier, Scene* scene, ShapeDescription* shape, const NxOgre::Pose& pose, ActorParams params = ""):
	  Trigger(meshIdentifier, scene, shape, pose, params)
	  {
	  }

	 void onEnter(Actor* a)
	 {
		int ize = 0;
	 }
};

class MyCharMoveController : public CharacterMovementVectorController
{
public:
	MyCharMoveController(float speed = 1.5){movespeed = speed;}
	float movespeed;
	void move(NxVec3 &out, NxVec3 &moveVector, NxQuat &direction, NxVec3 &g, float t, Character*)
	{
		NxVec3 dir = direction.rot(moveVector);
		out = ((dir * movespeed) + g) * t;		
	}
};

PlayerCharacter::PlayerCharacter(String name, Scene* scene, SceneManager* sceneManager, Vector3 dimensions, Vector3 startPosition, bool debug)
{
	mName = name;
	mPlayerDimensions = dimensions;
	mDebug = debug;
	mSceneManager = sceneManager;
	mScene = scene;
	CharacterParams cp;
	cp.setToDefault();
	cp.mDimensions = NxVec3(dimensions.x, dimensions.y, dimensions.z);
	cp.mType = CharacterParams::CT_Box;
	mPlayerCharacter = mScene->createCharacter(mName, startPosition, cp); 
	mPlayerCharacter->createNode();
	mPlayerNode = mPlayerCharacter->getNode();
	
	mCharacterController = 0;
	mCharacterController = new MyCharMoveController(2.0);
	mPlayerCharacter->setMovementVectorController(mCharacterController);

	if(mDebug)
	{
		mPlayerBoundDebugNode = mPlayerNode->createChildSceneNode(mName + "_BoundDebugNode");
		mPlayerBoundDebugNode->setScale(dimensions);
		mPlayerBoundDebugNode->attachObject(mSceneManager->createEntity(mName + "_BoundDebug", "cube.mesh"));
	}

	mCharacterMesh_AnimState = 0;
	mForceAnimChange = false;
	mWeapon = 0;
	mWeaponBound = 0;
	mWeaponBoundDebugNode = 0;
	mCharacterMeshEntity = 0;
	mWeaponEntity = 0;
	mWeaponAttachBone = 0;
    mWeaponNode = 0;
	mWeaponBoundNode = 0;
	mWeaponBoundDebugNode = 0;
	mPlayerBoundDebugNode = 0;
	mWeaponBound = 0;
	mWeaponNode2 = 0;
		
	this->setPriority(frameListenerPriority++);
	Root::getSingleton().addFrameListener(this);

	//mPlayerCharacter->setCharacterCollisionGroup(NX_CHARACTER_COL_GROUP_CHARACTER);
	//mScene->getNxScene()->setGroupCollisionFlag(NX_CHARACTER_COL_GROUP_CHARACTER, NX_CHARACTER_COL_GROUP_NON_COLLIDABLE,false);
}

FPSPlayer::FPSPlayer(Scene* scene, SceneManager* sceneManager, Vector3 dimensions, Vector3 startPosition, float cameraHeight, Camera* cam, bool debug)
: PlayerCharacter("FPSPlayerCharacter", scene, sceneManager, dimensions, startPosition, debug)
{
	mCameraHeight = cameraHeight - dimensions.y / 2.0;
	mCamera = cam;
	mPlayerNode->attachObject(cam); 
	
	thePlayer = this;
}

FPSPlayer::~FPSPlayer(void)
{
}

void PlayerCharacter::addPlayerMesh(String meshname, float meshscale, bool rotateModel)
{
	mCharacterMeshEntity = mSceneManager->createEntity(mName, meshname);
	mCharacterMeshEntity->setCastShadows(true);
	mMeshScale = meshscale;
	mCurrentAnimState = PLAYERANIMSTATE_STAND;
	mCharacterMesh_AnimState = mCharacterMeshEntity->getAnimationState(getAnimStateName(PLAYERANIMSTATE_STAND));
	mCharacterMesh_AnimState->setEnabled(true);
	mCharacterMesh_AnimState->setLoop(true);
	SceneNode* meshNode = mPlayerNode->createChildSceneNode(mName + "_MeshNode");
	meshNode->attachObject(mCharacterMeshEntity);
	meshNode->setPosition(0 ,-mPlayerDimensions.y / 2.0, 0);
	meshNode->setScale(meshscale, meshscale, meshscale);
	if(rotateModel)
	meshNode->rotate(Vector3::UNIT_Y, Degree(180.0));	
}

void FPSPlayer::addPlayerMesh(String meshname, float meshscale, bool rotateModel)
{
	mCharacterMeshEntity = mSceneManager->createEntity(mName, meshname);
	mCharacterMeshEntity->setCastShadows(false);
	mMeshScale = meshscale;
	mCurrentAnimState = PLAYERANIMSTATE_STAND;
	mCharacterMesh_AnimState = mCharacterMeshEntity->getAnimationState(getAnimStateName(PLAYERANIMSTATE_STAND));
	mCharacterMesh_AnimState->setEnabled(true);
	mCharacterMesh_AnimState->setLoop(true);
	SceneNode* FPSArmNode = mPlayerNode->createChildSceneNode(mName+"_ArmNode");
	FPSArmNode->attachObject(mCharacterMeshEntity); 
	FPSArmNode->setPosition(0,mCameraHeight,0);
	FPSArmNode->setScale(meshscale, meshscale, meshscale);
	if(rotateModel)
	FPSArmNode->rotate(Vector3::UNIT_Y, Degree(180.0));
}

void PlayerCharacter::addWeaponModel(String meshname, String attachBoneName)
{
	
	mWeaponEntity = mSceneManager->createEntity(mName + "_Weapon", meshname);
	mWeaponAttachBone = mCharacterMeshEntity->getSkeleton()->getBone(attachBoneName);
	mWeaponNode = mCharacterMeshEntity->getParentSceneNode()->createChildSceneNode(mName + "_WeaponNode");	
	mWeaponNode->attachObject(mWeaponEntity);

	AxisAlignedBox bb = mWeaponEntity->getBoundingBox();

	Vector3 max = bb.getMaximum();
	//max.z *= 2.0;
	Vector3 min = bb.getMinimum();
	Vector3 center = (min + max) / 2.0;
	//center.z = max.z;

	mWeaponBoundNode = mWeaponNode->createChildSceneNode(mName + "_WeaponBoundNode");
	mWeaponBoundNode->setPosition(center);
/*
	SceneNode* ArmNode2 = mPlayerNode->createChildSceneNode();
	ArmNode2->setPosition(0,mCameraHeight , -mPlayerDimensions.z);
	ArmNode2->rotate(Vector3::UNIT_Y, Degree(180.0));
	mWeaponNode2 = ArmNode2->createChildSceneNode();
	mWeaponBoundNode = mWeaponNode2->createChildSceneNode();
	mWeaponBoundNode->setPosition(center * mMeshScale );
*/	

	max *= mMeshScale;
	min *= mMeshScale;
	float dimX = max.x-min.x;
	float dimY = max.y-min.y;
	float dimZ = max.z-min.z;
	Vector3 dim(dimX, dimY, dimZ);

	mWeaponBound = new MyTrigger(mName + "_WeaponTrigger", mScene, new CubeShape(dimX, dimY, dimZ, "trigger: yes"), Vector3(0,0,0), "mass: 0, density: 0");
	mScene->addTrigger((Trigger*)mWeaponBound);
		
	if(mDebug)
	{
		mWeaponBoundDebugNode = mSceneManager->getRootSceneNode()->createChildSceneNode(mName+"_WeaponBoundDebugNode");
		mWeaponBoundDebugNode->setScale(dim);
		mWeaponBoundDebugNode->attachObject(mSceneManager->createEntity(mName+"_WeaponBoundDebug", "cube.mesh"));
	}

	//mWeaponBound->setCharacterCollisionGroup(NX_CHARACTER_COL_GROUP_NON_COLLIDABLE);	
}

void FPSPlayer::addWeaponModel(String meshname, String attachBoneName)
{
	
	mWeaponEntity = mSceneManager->createEntity(mName + "_Weapon", meshname);
	mWeaponEntity->setCastShadows(false);
	mWeaponAttachBone = mCharacterMeshEntity->getSkeleton()->getBone(attachBoneName);
	mWeaponNode = mCharacterMeshEntity->getParentSceneNode()->createChildSceneNode(mName+"_WeaponNode");	
	mWeaponNode->attachObject(mWeaponEntity);

	AxisAlignedBox bb = mWeaponEntity->getBoundingBox();

	Vector3 max = bb.getMaximum();
	//max.z *= 2.0;
	Vector3 min = bb.getMinimum();
	Vector3 center = (min + max) / 2.0;
	//center.z = max.z;

	//mWeaponBoundNode = mWeaponNode->createChildSceneNode();
	//mWeaponBoundNode->setPosition(center);

	SceneNode* ArmNode2 = mPlayerNode->createChildSceneNode(mName+"_ArmNode2");
	ArmNode2->setPosition(0,mCameraHeight , -mPlayerDimensions.z);
	ArmNode2->rotate(Vector3::UNIT_Y, Degree(180.0));
	mWeaponNode2 = ArmNode2->createChildSceneNode();
	mWeaponBoundNode = mWeaponNode2->createChildSceneNode(mName+"_WeaponBoundNode");
	mWeaponBoundNode->setPosition(center * mMeshScale );
	

	max *= mMeshScale;
	min *= mMeshScale;
	float dimX = max.x-min.x;
	float dimY = max.y-min.y;
	dimY *=3;
	float dimZ = max.z-min.z;
	Vector3 dim(dimX, dimY, dimZ);

//	mWeaponBound = mScene->createActor("FPSPlayerWeapon", new CubeShape(dimX, dimY, dimZ), Vector3(0,0,0), "kinematic: yes, mass: 10");
	mWeaponBound = new MyTrigger(mName + "_WeaponTrigger", mScene, new CubeShape(dimX, dimY, dimZ, "trigger: yes"), Vector3(0,0,0), "mass: 0, density: 0");
	mScene->addTrigger((Trigger*)mWeaponBound);
	
	if(mDebug)
	{
		mWeaponBoundDebugNode = mSceneManager->getRootSceneNode()->createChildSceneNode(mName+"_WeaponBoundDebugNode");
		mWeaponBoundDebugNode->setScale(dim);
		mWeaponBoundDebugNode->attachObject(mSceneManager->createEntity(mName+"_WeaponBoundDebug", "cube.mesh"));
	}

	mWeaponBound->setCharacterCollisionGroup(NX_CHARACTER_COL_GROUP_NON_COLLIDABLE);	
}

void PlayerCharacter::doAction(float dt)
{
	FPSPlayer* thePlayer = FPSPlayer::getPlayer();

	if(thePlayer)
	{
		Vector3 playerPos = thePlayer->getPosition();
		Vector3 myPos = mPlayerNode->getWorldPosition();
		Vector3 dir = playerPos - myPos;
		dir.y = 0;
		//if( dir.length() > 1.0)
		//{
			dir.normalise();
			Vector3 right = -dir.crossProduct(Vector3::UNIT_Y);
			Quaternion q; 			
			q.FromAxes(right,Vector3::UNIT_Y,dir);
			mPlayerCharacter->setDirection(q); 
			if(mCharacterController )
			((MyCharMoveController*)mCharacterController)->movespeed = 1.5;
			mPlayerCharacter->addMovement(Character::DR_StepRight);
		//}

	}
	if(mCharacterMeshEntity)
	setAnimState(PLAYERANIMSTATE_WALK);
}

bool PlayerCharacter::frameStarted(const FrameEvent& evt)
{
//	LogManager::getSingleton().logMessage("CHARACTERPLAYER_FRAMESTARTED");
	doAction(evt.timeSinceLastFrame);

	float dt = evt.timeSinceLastFrame;
		
	if(mCharacterMesh_AnimState)
	{
		float animspeed = 1.2;
		if(mCurrentAnimState == PLAYERANIMSTATE_WALK && !mRunning)
			animspeed /= 2;

		mCharacterMesh_AnimState->addTime(dt*animspeed);	
	}

	return true;
}

void FPSPlayer::processKeyInputs(InputReader* inputDevice)
{		
		mMoving = false;
		if (inputDevice->isKeyDown(KC_W))
		{ 
			mPlayerCharacter->addMovement(Character::DR_StepLeft); 
			mMoving = true;
		} 	
		if (inputDevice->isKeyDown(KC_S))
		{ 
			mPlayerCharacter->addMovement(Character::DR_StepRight);
			mMoving = true;
		} 
		if (inputDevice->isKeyDown(KC_A))
		{ 
			mPlayerCharacter->addMovement(Character::DR_Forward);
			mMoving = true;
		} 
		if (inputDevice->isKeyDown(KC_D))
		{ 
			mPlayerCharacter->addMovement(Character::DR_Backward); 
			mMoving = true;
		} 

		if (inputDevice->isKeyDown(KC_1))
		{
			mWeapon = 0;
			mForceAnimChange = true;
		}
		if (inputDevice->isKeyDown(KC_2))
		{
			mWeapon = 1;
			mForceAnimChange = true;
		}

		if (inputDevice->isKeyDown(KC_LSHIFT))
		{
			if(mCharacterController)
			((MyCharMoveController*)mCharacterController)->movespeed = 4.0;
			mRunning = true;
		}
		else
		{
			if(mCharacterController)
			((MyCharMoveController*)mCharacterController)->movespeed = 2.0;
			mRunning = false;
		}
}

void FPSPlayer::processMouseInputs(InputReader* inputDevice)
{
	if( inputDevice->getMouseButton( 0 ) )//attack
		setAnimState(PLAYERANIMSTATE_ATTACK);
		
	else if( inputDevice->getMouseButton( 1 ) )//block
		setAnimState(PLAYERANIMSTATE_BLOCK);

	Quaternion q; 
	q.FromAngleAxis( Radian(Degree(-inputDevice->getMouseRelX() * 0.13)), Vector3::UNIT_Y ); 
	q = mPlayerCharacter->getGlobalOrientation() * q; 
	mPlayerCharacter->setDirection(q); 

	mCamera->pitch(Radian(Degree(-inputDevice->getMouseRelY() * 0.13)));		
}

bool FPSPlayer::frameStarted(const FrameEvent& evt)
{
//	LogManager::getSingleton().logMessage("FPSPLAYER_FRAMESTARTED");

	float dt = evt.timeSinceLastFrame;
		
	if(mCharacterMesh_AnimState)
	{
		float animspeed = 1.2;
		if(mCurrentAnimState == PLAYERANIMSTATE_WALK && mRunning)
			animspeed *= 1.5;
		if(mCurrentAnimState == PLAYERANIMSTATE_ATTACK)
			animspeed *= 2.0;

		mCharacterMesh_AnimState->addTime(dt*animspeed);	
	}

	static float moveTime=0;
	
	float movecameraup = 0;
	if(mDebug)movecameraup = 20.0;

	if(mMoving)
	{
		setAnimState(PLAYERANIMSTATE_WALK);

		moveTime += dt;
		mCamera->setPosition(Vector3(0,mCameraHeight + movecameraup,0) + Vector3(0,0.009 * pow(sin(moveTime*5.0), 2),0));
	}
	else
	{
		if(Math::UnitRandom() > 0.98)
			setAnimState(PLAYERANIMSTATE_IDLE);
		else
			setAnimState(PLAYERANIMSTATE_STAND);

		moveTime = 0;
		mCamera->setPosition(Vector3(0,mCameraHeight + movecameraup,0));
	}

	if(mCharacterMeshEntity)
	this->mCharacterMeshEntity->_updateAnimation();

	if(mWeaponEntity)
	{
		mWeaponNode->setPosition(mWeaponAttachBone->getWorldPosition());
		mWeaponNode->setOrientation(mWeaponAttachBone->getWorldOrientation());

		mWeaponNode2->setPosition(mWeaponAttachBone->getWorldPosition() * mMeshScale);
		mWeaponNode2->setOrientation(mWeaponAttachBone->getWorldOrientation());

		Vector3 translation = mWeaponBoundNode->getWorldPosition();
		Quaternion orientation = mWeaponBoundNode->getWorldOrientation();
		
		mWeaponBound->setGlobalOrientation(orientation);		
		mWeaponBound->setGlobalPosition(translation);

		//mWeaponBound->setGlobalPosition(translation);		
		//mWeaponBound->setGlobalOrientation(orientation);

		if(mDebug)
		{
			mWeaponBoundDebugNode->setPosition(translation);
			mWeaponBoundDebugNode->setOrientation(orientation);
		}
	}

	return true;
}

String FPSPlayer::getAnimStateName(PlayerAnimState as, int option)
{
	switch (as)
	{
		case PLAYERANIMSTATE_IDLE:
		{
			if(mWeapon < 1)
				return "sword_idle";
			else
				return "magic_idle";
		}
		case PLAYERANIMSTATE_STAND:
		{
			if(mWeapon < 1)
				return "sword_stand";
			else
				return "magic_stand";
		}
		case PLAYERANIMSTATE_BLOCK:
		{
				return "block";			
		}
		case PLAYERANIMSTATE_WALK:
		{
				return "walk";			
		}
		case PLAYERANIMSTATE_ATTACK:
		{
			if(option)
			{
				if(mWeapon < 1)
					return "attack1";
				else
					return "magic_attack";
			}
			else
			{
				if(mWeapon < 1)
					return "attack2";
				else
					return "magic_attack";
			}
		}
	}
	return "";
}

void FPSPlayer::setAnimState(PlayerAnimState as)
{
	if(!mCharacterMeshEntity)
		return;

	switch(as)	
	{
		case PLAYERANIMSTATE_STAND:
		{
			if(mCurrentAnimState != PLAYERANIMSTATE_STAND || mForceAnimChange)
			{
				//wait until anim finishes
				if(mCharacterMesh_AnimState->getTimePosition() >=
					mCharacterMesh_AnimState->getLength() || mCharacterMesh_AnimState->getLoop())
				{
					mCurrentAnimState = PLAYERANIMSTATE_STAND;
					mCharacterMesh_AnimState->setEnabled(false);
					mCharacterMesh_AnimState = mCharacterMeshEntity->getAnimationState(getAnimStateName(PLAYERANIMSTATE_STAND));
					mCharacterMesh_AnimState->setLoop(true);
					mCharacterMesh_AnimState->setEnabled(true);
				}
			}
			break;
		}
		case PLAYERANIMSTATE_IDLE:
		{
			if(mCurrentAnimState != PLAYERANIMSTATE_IDLE || mForceAnimChange)
			{
				//wait until anim finishes
				if(mCharacterMesh_AnimState->getTimePosition() >=
					mCharacterMesh_AnimState->getLength() || mCharacterMesh_AnimState->getLoop() || mCurrentAnimState == PLAYERANIMSTATE_IDLE)
				{
					mCurrentAnimState = PLAYERANIMSTATE_IDLE;
					mCharacterMesh_AnimState->setEnabled(false);
					mCharacterMesh_AnimState = mCharacterMeshEntity->getAnimationState(getAnimStateName(PLAYERANIMSTATE_IDLE));
					mCharacterMesh_AnimState->setTimePosition(0);
					mCharacterMesh_AnimState->setLoop(false);
					mCharacterMesh_AnimState->setEnabled(true);
				}
			}
			break;
		}
		case PLAYERANIMSTATE_WALK:
		{
			if(mCurrentAnimState != PLAYERANIMSTATE_WALK)
			{
				if(mCurrentAnimState == PLAYERANIMSTATE_IDLE || mCurrentAnimState == PLAYERANIMSTATE_STAND)
				{
					mCurrentAnimState = PLAYERANIMSTATE_WALK;
					mCharacterMesh_AnimState->setEnabled(false);
					mCharacterMesh_AnimState = mCharacterMeshEntity->getAnimationState(getAnimStateName(PLAYERANIMSTATE_WALK));
					mCharacterMesh_AnimState->setLoop(true);
					mCharacterMesh_AnimState->setEnabled(true);
				}
			}
			break;
		}
		case PLAYERANIMSTATE_ATTACK:
		{
			if(mCharacterMesh_AnimState->getTimePosition() >=
					mCharacterMesh_AnimState->getLength() || mCharacterMesh_AnimState->getLoop() || mCurrentAnimState == PLAYERANIMSTATE_IDLE)
			{
			    int attackType = 0;
				if(Math::UnitRandom() > 0.5)
					 attackType = 1;

				mCurrentAnimState = PLAYERANIMSTATE_ATTACK;
				mCharacterMesh_AnimState->setEnabled(false);
				mCharacterMesh_AnimState = mCharacterMeshEntity->getAnimationState(getAnimStateName(PLAYERANIMSTATE_ATTACK, attackType));
				mCharacterMesh_AnimState->setTimePosition(0);
				mCharacterMesh_AnimState->setLoop(false);
				mCharacterMesh_AnimState->setEnabled(true);
			}
			break;
		}		
		case PLAYERANIMSTATE_BLOCK:
		{
			if(mCharacterMesh_AnimState->getTimePosition() >=
					mCharacterMesh_AnimState->getLength() || mCharacterMesh_AnimState->getLoop() || mCurrentAnimState == PLAYERANIMSTATE_IDLE)
			{
				mCurrentAnimState = PLAYERANIMSTATE_BLOCK;
				mCharacterMesh_AnimState->setEnabled(false);
				mCharacterMesh_AnimState = mCharacterMeshEntity->getAnimationState(getAnimStateName(PLAYERANIMSTATE_BLOCK));
				mCharacterMesh_AnimState->setTimePosition(0);
				mCharacterMesh_AnimState->setLoop(false);
				mCharacterMesh_AnimState->setEnabled(true);
			}
			break;
		}
	}

	if(mForceAnimChange)
		mForceAnimChange = false;
}
